#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>

// Η δομή SuperString όπως είχε δοθεί και στην προηγούμενη
// εργασία (της Β΄ Ανακεφαλαίωσης)
typedef struct _super_string {
    char *text;
    int size;
    int baseSize;
} SuperString;

/*
 * Ανάγνωση ενός (του μοναδικού) SuperString που περιέχεται
 * στο αρχείο με όνομα (ή διαδρομή/path) fileName.
 * Επιστρέφει ένα νέο SuperString όπως ήταν στο αρχείο.
 */
SuperString *ssReadFromFile(char *fileName) {
    // Ανοίγουμε το αρχείο για ανάγνωση σε binary μορφή
    FILE *fin = fopen(fileName, "rb");
    // Ελέγχουμε εάν βρέθηκε το αρχείο και άνοιξε
    if (fin == NULL) {
        return NULL;
    }

    // Δεσμεύουμε μνήμη για τη δομή
    SuperString *s = (SuperString *)malloc(sizeof(SuperString));

    // Διαβάζουμε από το αρχείο τη δομή (χωρίς δηλαδή το κείμενο)
    if (fread(s, sizeof(SuperString), 1, fin) != 1) {
        // εάν δεν επιτύχει, τότε αποδεσμεύουμε ότι έχουμε σε resources
        fclose(fin);
        return NULL;
    }

    // Δεσμεύουμε μνήμη βάσει του μεγέθους που αναγράφει η δομή στο αρχείο
    s->text = (char *)malloc(s->size);
    // ελέγχουμε εάν βρέθηκε η μνήμη
    if (s->text == NULL) {
        // εάν όχι, τότε αποδεσμεύουμε ότι έχουμε σε resources
        free(s);
        fclose(fin);
        return NULL;
    }

    // Διαβάζουμε από το αρχέιο (έναν χαρακτήρα λιγότερο καθώς το 0 που
    // τερματίζει τα strings δεν περιλαμβάνεται στο αρχείο)
    int readBytes = fread(s->text, sizeof(char), s->size-1, fin);
    // ελέγχουμε εάν διαβάστηκε όλη η πληροφορία
    // (δηλαδή και τα s->size-1 bytes/χαρακτήρες)
    if (readBytes != s->size-1) {
        // εάν δεν επιτύχει, τότε αποδεσμεύουμε ότι έχουμε σε resources
        free(s->text);
        free(s);
        fclose(fin);
        return NULL;
    }
    // τοποθετούμε στο τέλος του κειμένου τον μηδενικό τερματικό χαρακτήρα
    s->text[s->size-1] = 0;

    // κλείνουμε το αρχείο
    fclose(fin);
    // επιστρέφουμε τη δομή
    return s;
}

/*
 * Εγγραφή μίας μόνης δομής SuperString στο αρχείο με
 * όνομα fileName
 */
bool ssWriteToFile(SuperString *ss, char *fileName) {
    // ανοίγουμε το αρχείο για εγγραφή (ή δημιουργία) σε binary mode
    FILE *fout = fopen(fileName, "w+b");
    // ελέγχουμε εάν όντως άνοιξε
    if (fout == NULL) {
        return false;
    }

    // Κανονικά εδώ για οικονομία χώρου στον δίσκο πρέπει να καλέσουμε
    // την ssTruncate ώστε να μην μεταφερθεί η αχρησιμοποίηση μνήμη στον δίσκο

    // πρώτα γράφεται στον δίσκο η δομή αρχικά
    if ( fwrite(ss, sizeof(SuperString), 1, fout) != 1) {
        // αν αποτύχει η εγγραφή τερματίζει η συνάρτηση κατάλληλα
        fclose(fout);
        return false;
    }

    // ακολούθως γράφεται στον δίσκο το ίδιο το κείμενο χωρίς τον τερματικό
    // μηδενικό χαρακτήρα (για οικονομία, έστω και αυτού του ενός byte)
    if ( fwrite(ss->text, sizeof(char), ss->size - 1, fout) != ss->size - 1 ) {
        // αν αποτύχει η εγγραφή τερματίζει η συνάρτηση κατάλληλα
        fclose(fout);
        return false;
    }

    // κλείνουμε το αρχείο - ώστε να γραφτί και ο buffer στον δίσκο
    fclose(fout);

    // τερματίζεται επιτυχώς η συνάρτηση
    return true;
}

int main() {
    SuperString s;
    s.text = "Dokimastiko";
    s.size = 9;
    s.baseSize = 15;

    // Υπόδειγμα εγγραφής στο αρχείο
//    ssWriteToFile(&s, "../sample-output.sup");

    // Υπόδειγμα ανάγνωση από το αρχείο
    SuperString *s2 = ssReadFromFile("../sample-output.sup");
    if (s2 != NULL) {
        printf("%s\n%d\n%d\n", s2->text, s2->size, s2->baseSize);
    } else {
        printf("ERROR!");
    }

    return 0;
}
